--------------------------------------------------------------------------------
-- LEON3 Usecase JRZ VECS - FH Technikum Wien (c) 2013/2014
-- File: assertion_collector.vhd                                                      --
-- Author: matthias                                                          --
-- Created on: 13.2.2014                                                      --
--                                                                            --
-- Keeps book about failing/finishing assertions, requests irq upon single failure;
-- raises new one if acknowlegded in this peripheral
-- can cause th system to reset if configured this way (set resp, bit in coll register)
-- apb bus peripheral
-- based on grpreg
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library grlib;
use grlib.amba.all;
use grlib.stdlib.all;
use grlib.devices.all;
library gaisler;
use gaisler.misc.all;

library work;
use work.assertion_checkers_p.all;

--pragma translate_off
use std.textio.all;
--pragma translate_on

entity assertion_collector is
  generic (
    pindex  : integer               := 0;
    paddr   : integer               := 0;
    pmask   : integer               := 16#fff#;
    pirq    : integer               := 0;
    nbits   : integer range 1 to 64 := 32;
    rstval  : integer               := 0;
    rstval2 : integer               := 0;
    extrst  : integer               := 0
    );
  port (
    rst      : in  std_ulogic;
    clk      : in  std_ulogic;
    apbi     : in  apb_slv_in_type;
    apbo     : out apb_slv_out_type;
    self_reset_o :out std_logic;
    fatal_fail_o :out std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
    refeed_self_reset: in std_logic_vector (1 downto 0);
    ass_stat : in  assertion_status_t;
    resval   : in  std_logic_vector(nbits-1 downto 0) := (others => '0')
    );
end;

architecture rtl of assertion_collector is

  constant REVISION            : integer := 0;
  constant VENDOR_NONE         : integer := 16#de#;
  constant ASSERTION_COLLECTOR : integer := 16#ad#;
  

  constant pconfig : apb_config_type := (
    0 => ahb_device_reg (VENDOR_NONE, ASSERTION_COLLECTOR, 0, REVISION, pirq),
    1 => apb_iobar(paddr, pmask));

  type registers is record
    coll : std_logic_vector(nbits-1 downto 0);
    cfg  : std_logic_vector(nbits-1 downto 0);
    pend : std_logic_vector(nbits-1 downto 0);
  end record;

  signal failed_edge1_r : std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
  signal failed_edge2_r : std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
  signal fatal_fail, fatal_fail_r :std_logic_vector(ASS_STAT_WIDTH-1 downto 0);

  signal r, rin : registers;
	signal self_reset :std_logic := '0';--simulation help,
																			--this fpga initializes everything to '0'
	signal self_reset_r : std_logic;
  
begin

  comb : process(ass_stat, rst, r, apbi, resval, failed_edge1_r, failed_edge2_r, self_reset_r, fatal_fail_r, refeed_self_reset)
    variable readdata : std_logic_vector(31 downto 0);
    variable v        : registers;
  begin
		self_reset <= self_reset_r;
		fatal_fail <= fatal_fail_r;
    v := r;
-- read register

    readdata := (others => '0');
    case apbi.paddr(4 downto 2) is
      when "000" =>
        readdata(nbits-1 downto 0) := r.coll;
      when "001" =>
        readdata(nbits-1 downto 0) := r.cfg;
      when "010" =>
        readdata(nbits-1 downto 0) := r.pend;
      when others =>
    end case;

-- write registers

    if (apbi.psel(pindex) and apbi.penable and apbi.pwrite) = '1' then
      case apbi.paddr(4 downto 2) is
        when "000" =>                   --self_reset regs
          v.coll := apbi.pwdata;
        when "001" =>                   --cfg regs
          v.cfg := apbi.pwdata;
        when "010" =>                   --irq pending
          for i in 0 to nbits-1 loop
            if apbi.pwdata(i) = '1' then
              v.pend(i) := '0';
            end if;
          end loop;
        when others =>
      end case;
    end if;

    --

    --for i in 0 to ASS_STAT_WIDTH-1 loop
    --  v.coll(i) := ass_stat(i).failed;
    --end loop;
--         for i in 0 to 7 loop
--        v.coll(i*3) := ass_stat(i).pending;
--        v.coll((i*3)+1) := ass_stat(i).done;
--        v.coll((i*3)+2) := ass_stat(i).failed;
--      end loop;

    --generate an interrupt upon a rising edge of the first failed assertion
    apbo.pirq <= (others => '0');
    for i in 0 to ASS_STAT_WIDTH-1 loop
      if(failed_edge2_r(i) = '0' and failed_edge1_r(i) = '1') then
        if(v.cfg(i) = '1' and v.pend(i) = '0') then
					if(v.coll(i) = '1') then --check if self reset must be triggered
						self_reset <= '1';
						fatal_fail(i) <= '1';
					else --else tree necessary?
						self_reset <= '0';
						fatal_fail(i) <= '0';
					end if;
          v.pend(i)       := '1';
          apbo.pirq(pirq) <= '1';
        end if;
      end if;
      
    end loop;
    --
    if rst = '0' then
			if(self_reset_r /= '1') then --do not reset self reset if it triggered
																	-- the reset
				fatal_fail <= (others => '0');
				self_reset <= '0';
			end if;
			
      if extrst = 0 then
        v.coll := conv_std_logic_vector(rstval, nbits);
        v.cfg  := conv_std_logic_vector(rstval, nbits);
        v.pend := conv_std_logic_vector(rstval, nbits);
      else
        v.coll := resval;
        v.cfg  := resval;
        v.pend := resval;
      end if;
    end if;

		if refeed_self_reset = "11" then --ok self reset has happened
																			--reset self reset signal
			fatal_fail <= (others => '0');
			self_reset <= '0';
		end if;
    
    rin <= v;

    apbo.prdata <= readdata;            -- drive apb read bus


  end process;

  apbo.pindex  <= pindex;
  apbo.pconfig <= pconfig;
	self_reset_o <= self_reset_r;
	fatal_fail_o <= fatal_fail_r;
-- registers

  regs : process(clk)
  begin
    if rising_edge(clk) then
      r <= rin;
      self_reset_r <= self_reset;
      fatal_fail_r <= fatal_fail;

      for i in 0 to ASS_STAT_WIDTH-1 loop
        failed_edge1_r(i) <= ass_stat(i).failed;
      end loop;

      failed_edge2_r <= failed_edge1_r;
    end if;
  end process;

-- boot message

-- pragma translate_off
  bootmsg : report_version
    generic map ("assertion_collectorg" & tost(pindex) &
                 ": " & tost(nbits) & "-bit ASSERTION_COLLECTOR Unit rev " &
                 tost(REVISION) & " irq " & tost(pirq));
-- pragma translate_on

end;
