--------------------------------------------------------------------------------
-- LEON3 Usecase JRZ VECS - FH Technikum Wien (c) 2013/2014
-- File:i2c_liveness.vhd                                                      --
-- Author: matthias                                                          --
-- Created on: 4.3.2014                                                      --
--                                                                            --
-- checks for in general correct i2c behaviour; (minimal amount of clock cylces (address,data), start/stop conditions)
-- property scl_liveness(const min; const max) is always (
--	  (fell(i2ci.sda) and i2ci.scl = '1') -> (
--		  {i2ci.sda = '0'}[*min to max] |->  (
--		    {fell(i2ci.scl)} |=> ( {{{{i2ci.scl = prev(i2ci.scl)}[*min to max]; (rose(i2ci.scl) or fell(i2ci.scl))}}[*18]}[*100])
--			    abort ((rose(i2ci.sda) and i2ci.scl = '1') or (fell(i2ci.sda) and i2ci.scl = '1'))
--		  ) 	  
--	  )
--	);
--------------------------------------------------------------------------------


library ieee;
	use ieee.std_logic_1164.all;
	use IEEE.std_logic_arith.all;
library work;
	use work.assertion_checkers_p.all;
	
entity i2c_liveness is
	port (
		clk_i     : in std_logic;
		reset_i   : in std_logic;
		-- assertion inputs --
		scl :in std_logic;
		sda :in std_logic;
	
			
		-- assertion status --
		pending_o : out std_logic;
		done_o    : out std_logic;
		failed_o  : out std_logic
	
	);

end entity i2c_liveness;


architecture rtl of i2c_liveness is

  signal  scl_r, scl_r_2	: std_logic;
  signal  sda_r, sda_r_2	: std_logic;
 


  type state_t is (WAIT_S,OBEY_SDA_LOW_S,PRE_CLOCKING_S,CLOCKING_S);
  signal state_act, state_nxt : state_t;
  
  
  signal pending, pending_r :std_logic;
  signal failed, failed_r :std_logic;
  signal done, done_r :std_logic;
  
  --wait counter
  signal reset_0_n :std_logic;
  signal cnt_enable_0 :std_logic;
  signal lower_thr_0 :std_logic;
  signal upper_thr_0 :std_logic;
  
  signal reset_1_n :std_logic;
  signal cnt_enable_1 :std_logic;
  signal lower_thr_1 :std_logic;
  signal upper_thr_1 :std_logic;
  
begin

  wait_counter_0 : wait_counter
	generic map(
	  G_LOWER_BOUND  =>100,
	  G_UPPER_BOUND => 15000 
	)
	port map(
		clk_i  =>  clk_i,
		reset_i  =>  reset_0_n,	
		enable_i => cnt_enable_0,
		lower_thr_o => lower_thr_0,
		upper_thr_o => upper_thr_0	
	);

  wait_counter_1 : wait_counter
	generic map(
	  G_LOWER_BOUND  =>35,--(18*2)-1
	  G_UPPER_BOUND => 35 
	)
	port map(
		clk_i  =>  clk_i,
		reset_i  =>  reset_1_n,	
		enable_i => cnt_enable_1,
		lower_thr_o => lower_thr_1,
		upper_thr_o => upper_thr_1	
	);
	

  p_check : process(clk_i,reset_i)
  begin
    if(reset_i = '0') then
      state_act <= WAIT_S;
    
     
      scl_r <= '0';
      scl_r_2 <= '0';
      
      sda_r <= '0';
      sda_r_2 <= '0';

      
      done_r <= '0';
      pending_r <= '0';
      failed_r <= '0';
    
    elsif rising_edge(clk_i) then
      
      state_act <= state_nxt;
      
      scl_r <= scl;
      scl_r_2 <= scl_r;
      
      sda_r <= sda;
      sda_r_2 <= sda_r;
      
      
      done_r <= done;
      pending_r <= pending;
      failed_r <= failed;
    
    end if;
  
  
  end process;
  
  
  p_check_comb: process(done_r, pending_r, failed_r, sda_r, sda_r_2, scl_r, scl_r_2, upper_thr_0, lower_thr_0, upper_thr_1, lower_thr_1, state_act) is
  begin
    state_nxt <= state_act;
    done <= done_r;
    pending <= pending_r;
    failed <= failed_r;
    reset_0_n <= '1';
    reset_1_n <= '1';
    cnt_enable_0 <= '0';
    cnt_enable_1 <= '0';
    
    case state_act is
      when WAIT_S =>
	pending <= '0';
	done <= '0';
	failed <= '0';
	
	--start condition found
	if((sda_r_2 = '1' and sda_r = '0') and scl_r = '1') then
	  reset_0_n <= '0';
	  pending <= '1';
	  state_nxt <= OBEY_SDA_LOW_S;
	end if;
	
      when OBEY_SDA_LOW_S =>
        --may be transitioningt from CLOCKING_S
	done <= '0'; --may be transitioningt from CLOCKING_S
	pending <= '0';
	cnt_enable_0 <= '1';
	--fail
	if ((sda_r = '1' and lower_thr_0 = '0') or (upper_thr_0 = '1')) then
	  failed <= '1';
	  cnt_enable_0 <= '0';
	  --resets counter 0 upon start condition detection
	  state_nxt <= WAIT_S;
	end if;
	
	--next: clocking starts
	if(scl_r_2 = '0' and scl_r = '1') then
	  --reset message counter
	  reset_1_n <= '0';
	  state_nxt <= PRE_CLOCKING_S;
	end if;
	
      when PRE_CLOCKING_S =>
	  cnt_enable_0 <= '0';
	  reset_0_n <= '0';
	  state_nxt <= CLOCKING_S;
	
      when CLOCKING_S => 
	    cnt_enable_0 <= '1'; --scl clock counter
	
	    --scl clock event
	   if( (scl_r xor scl_r_2) = '1' ) then
	      cnt_enable_1 <= '1';--address, data or data, data i2c message counter
	      state_nxt <= PRE_CLOCKING_S;
	   end if;
	
	   --two messages have been transmitted, reset counter_1
	   if lower_thr_1 = '1' then
	    reset_1_n <= '0';
	    cnt_enable_1 <= '0';
	   end if;
	
	    --fail
	    if( ((scl_r xor scl_r_2)) = '1' and lower_thr_0 = '0') or (upper_thr_0 = '1') then
	      failed <= '1';
	      cnt_enable_0 <= '0';
	      state_nxt <= WAIT_S;
	    end if;
	   
	   --stop condition detected
	    if (sda_r_2 = '0' and sda_r = '1') and scl_r = '1'  then
	      pending <= '0';
	      done <= '1';
	      state_nxt <= WAIT_S;
	    end if;
	   
	   --start retransmission found, check fr lower_thr_1 here as well, check counter module
	    if((sda_r_2 = '1' and sda_r = '0') and scl_r = '1') then
	      reset_0_n <= '0';
	      cnt_enable_0 <= '0';
	      pending <= '0';
	      done <= '1';
	      state_nxt <= OBEY_SDA_LOW_S;
	    end if;
	   
	   	   
    
      when others => null;
    end case;
  

  
  end process;

  done_o <= done or done_r;
  pending_o <= pending or pending_r;
  failed_o <= failed or failed_r;
  

end architecture rtl;
