--------------------------------------------------------------------------------
-- LEON3 Usecase JRZ VECS - FH Technikum Wien (c) 2013/2014
-- File: fault_injection_button_sync.vhd                                      --
-- Author: christian                                                          --
-- Created on: 1.10.2013                                                      --
--                                                                            --
-- Synchronizes and debounces a specified number of inputs (buttons)          --
-- All input signals are considered to be asynchronous and sampled by two     --
-- register stages. The sync'd signal is transferred into the output register --
-- all (G_DEBOUNCE_COUNTER_WIDTH-1)^2 clock cycles.                           --
--	                                                                          --
-- Input signals for which the corresponding bit in the G_SINGLE_SHOT_MASK    --
-- is set are cleared after 1 clock cycle and can only activated again by     --
-- releasing the corresponding button for at least one sample cycle before    --
-- pressing it again.                                                         --
--                                                                            --
-- The signal value for a pressed button is taken from fault_injection_p      --
-- ('0' for the Cyclone V SoC Development Board                               --
--------------------------------------------------------------------------------

library ieee;
    use ieee.std_logic_1164.all;
    use ieee.std_logic_arith.all;

library work;
    use work.fault_injection_unit_p.all;
    
entity fault_injection_button_sync is
    generic (
        G_DEBOUNCE_COUNTER_WIDTH : natural;
        G_DEBOUNCE_SIGNAL_WIDTH  : natural;
        G_SINGLE_SHOT_MASK       : natural
    );
    port (
        clk_i     : in std_logic;
        reset_n_i : in std_logic;
        async_i   : in std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0);
        sync_o    : out std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0)
    );
end entity;

architecture rtl of fault_injection_button_sync is

    constant C_SINGLE_SHOT_MASK : unsigned(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0) := conv_unsigned(G_SINGLE_SHOT_MASK,G_DEBOUNCE_SIGNAL_WIDTH);
	constant C_OUTPUT_MASK      : std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0) := (others => (C_BUTTON_PRESSED));
    
    type debounce_reg_t is array(natural range <>) of std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0);
    
    signal s_trigger_cnt   : unsigned(G_DEBOUNCE_COUNTER_WIDTH-1 downto 0);
    signal s_debounce_regs : debounce_reg_t(1 downto 0);
	
    signal s_output_reg0   : std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0);
	signal s_output_reg1   : std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0);
	
	signal s_output_enable : std_logic_vector(G_DEBOUNCE_SIGNAL_WIDTH-1 downto 0);

begin

    -- simple assignments --
    gen_output : for i in 0 to G_DEBOUNCE_SIGNAL_WIDTH-1 generate
		sync_o(i) <= s_output_reg1(i) when s_output_enable(i) = '1' else not(C_BUTTON_PRESSED);
	end generate;
	

    p_sync_debounce: process(clk_i, reset_n_i)

    begin
        if reset_n_i = '0' then
            s_debounce_regs <= (others => (others => not(C_BUTTON_PRESSED)));
            s_output_reg0   <= (others => not(C_BUTTON_PRESSED));
			s_output_reg1   <= (others => not(C_BUTTON_PRESSED));
			s_output_enable <= (others => '0');
            s_trigger_cnt   <= (others => '0');
        elsif rising_edge(clk_i) then
            		
			s_output_enable <= not(std_logic_vector(C_SINGLE_SHOT_MASK));
					
            -- single shot signal default --
            for i in 0 to G_DEBOUNCE_SIGNAL_WIDTH-1 loop
                
                if C_SINGLE_SHOT_MASK(i) = '1' and (s_output_reg0(i) = C_BUTTON_PRESSED) and (s_output_reg1(i) = not(C_BUTTON_PRESSED)) then -- and s_output_enable(i) = '0' then
                    s_output_enable(i) <= '1';
                end if;
                
            end loop;
            
            -- synchronizing the trigger signals --
            
            s_debounce_regs(0) <= async_i;
            s_debounce_regs(1) <= s_debounce_regs(0);
            
            -- debouncing the trigger signals --
            
            if s_trigger_cnt(G_DEBOUNCE_COUNTER_WIDTH-1) = '1' then
			
				--assert false report "Read syncd button values." severity note;
				
				s_output_reg0   <= s_debounce_regs(1);
				s_output_reg1   <= s_output_reg0;
				
				s_trigger_cnt   <= (others => '0');
				
			else
				s_trigger_cnt <= s_trigger_cnt + '1';
			end if;
            
        end if;
    end process p_sync_debounce;
    
    

end rtl;
