--------------------------------------------------------------------------------
-- LEON3 Usecase JRZ VECS - FH Technikum Wien (c) 2013/2014
-- File: fault_injection_unit.vhd                                             --
-- Author: christian                                                          --
-- Created on: 1.10.2013                                                      --
--                                                                            --
-- trigger_i multiplexes data_o between                                       --
-- : data_i                                                                   --
-- : data_i <G_OPERATION> <G_FI_DATA/s_lfsr_data>                             --
--                                                                            --
-- Configuration options:                                                     --
-- G_ASYNC_TRIGGER           : Is the trigger signal async?                   --
--                             Should a synchronizer be implemented?          --  
-- G_DEBOUNCE_COUNTER_WIDTH  : Synchronizer/Debouncer period (n-1)^2          --
-- G_OPERATION               : Which operation shall be applied for FI        --
-- G_INJECT_SINGLE_SHOT      : Synchronizer/Debouncer single shot             --
-- G_RANDOMIZE               : Shall an LFSR be implemented?                  --
-- G_DATA_WIDTH              : Width of the data_i and data_o signal          --
-- G_FI_DATA                 : Static FI data or LFSR seed                    --
-- G_LFSR_POLY               : LFSR polynomial                                --
-- G_LFSR_WIDTH              : Width of the LFSR                              --
--                             FI data are the lower N bits of the LFSR       --
--                             Therefore, G_LFSR_WIDTH >= G_DATA_WIDTH        --
--                                                                            --
-- Trigger active level (C_BUTTON_PRESSED) is taken from fault_injection_p    --
--------------------------------------------------------------------------------

library ieee;
    use ieee.std_logic_1164.all;
    use ieee.std_logic_arith.all;
library work;
    use work.fault_injection_p.all;

entity fault_injection_unit is
    generic(
        G_ASYNC_TRIGGER           : boolean;
		G_DEBOUNCE_COUNTER_WIDTH  : natural range 1 to natural'right := 1;
        G_OPERATION               : t_operation := OP_OVW;
        G_INJECT_SINGLE_SHOT      : boolean := true;
        G_RANDOMIZE               : boolean := false;
        G_DATA_WIDTH              : natural := 8;
        G_FI_DATA                 : integer := 16#FF#;
        G_LFSR_POLY               : integer := 16#71#;
		G_LFSR_WIDTH              : integer := 8
    );
    port(
        clk_i                : in  std_logic;
        reset_n_i            : in  std_logic;
        trigger_i            : in  std_logic;
    
        data_i              : in  std_logic_vector(G_DATA_WIDTH-1 downto 0);
        data_o              : out std_logic_vector(G_DATA_WIDTH-1 downto 0)
    );
end entity;

architecture rtl of fault_injection_unit is
    
    signal s_trigger_syncd           : std_logic := not(C_BUTTON_PRESSED);
    signal s_enable_fi               : boolean := false;
    signal s_lfsr_data               : std_logic_vector(G_LFSR_WIDTH-1 downto 0) := (others => '0');
    
	constant C_LFSR_FEEDBACK : std_logic_vector(G_LFSR_WIDTH-1 downto 0) := std_logic_vector(conv_unsigned(G_LFSR_POLY,(s_lfsr_data'length)));

begin

    gen_standalone_sync : if G_ASYNC_TRIGGER = true generate
    
       -- s_inject_single_shot <= 1 when G_INJECT_SINGLE_SHOT else 0;
        
        button_sync_i : fault_injection_button_sync
            generic map(
                G_DEBOUNCE_COUNTER_WIDTH => G_DEBOUNCE_COUNTER_WIDTH,
                G_DEBOUNCE_SIGNAL_WIDTH  => 1,
                G_SINGLE_SHOT_MASK       => bool2nat(G_INJECT_SINGLE_SHOT)
            )
            port map(
                clk_i      => clk_i,
                reset_n_i  => reset_n_i,
                async_i(0) => trigger_i,
                sync_o(0)  => s_trigger_syncd
            );

    end generate;
    
    p_mux_trigger: process(s_trigger_syncd,trigger_i)
        variable v_trigger : std_logic;
    begin
        
        if G_ASYNC_TRIGGER then
            v_trigger := s_trigger_syncd;
        else
            v_trigger := trigger_i;
        end if;
        
        if (v_trigger = C_BUTTON_PRESSED) then
		    s_enable_fi <= true;
		else
		    s_enable_fi <= false;
		end if;
    
    end process p_mux_trigger;
    
    
    
	gen_lfsr: if G_RANDOMIZE generate
		p_lfsr : process (clk_i, reset_n_i)
			variable v_next_lfsr_value : std_logic_vector((s_lfsr_data'length)-1 downto 0);
		begin
				if reset_n_i = '0' then
					s_lfsr_data <= std_logic_vector(conv_unsigned(G_FI_DATA,(s_lfsr_data'length)));
				elsif rising_edge(clk_i) then
				
					s_lfsr_data((s_lfsr_data'length)-1 downto 1)  <= s_lfsr_data((s_lfsr_data'length)-2 downto 0);
					s_lfsr_data(0)                                <= s_lfsr_data((s_lfsr_data'length)-1);
					
					-- generate taps --
					for i in 1 to ((s_lfsr_data'length)-1) loop
						if(C_LFSR_FEEDBACK(i) = '1') then
							s_lfsr_data(i) <= s_lfsr_data(i-1) xor s_lfsr_data((s_lfsr_data'length)-1);
						end if;
					end loop;
					
				end if;
		end process p_lfsr;
    end generate;
	
    p_fimux : process (data_i,s_trigger_syncd,s_lfsr_data,s_enable_fi)
        variable v_fi_data   : std_logic_vector(G_DATA_WIDTH-1 downto 0);
    begin
        if G_RANDOMIZE then
            v_fi_data := s_lfsr_data(G_DATA_WIDTH-1 downto 0);
        else
            v_fi_data := std_logic_vector(conv_unsigned(G_FI_DATA,G_DATA_WIDTH));
		end if;

        if s_enable_fi then
			assert false report "Injecting Error" severity note;
            case G_OPERATION is
				when OP_AND =>
				   data_o <= data_i and v_fi_data;
				when OP_OR  =>
				   data_o <= data_i or v_fi_data;
				when OP_XOR =>
				   data_o <= data_i xor v_fi_data;
				when OP_OVW =>
					data_o <= v_fi_data;
				when others =>
					null;
			end case;
        else
          data_o <= data_i;
        end if;
    end process p_fimux;

end rtl;
