-----------------------------------------------------------------------------
--  LEON3 Demonstration design
--  Copyright (C) 2004 Jiri Gaisler, Gaisler Research
------------------------------------------------------------------------------
------------------------------------------------------------------------------
--  This file is a part of the GRLIB VHDL IP LIBRARY
--  Copyright (C) 2003 - 2008, Gaisler Research
--  Copyright (C) 2008 - 2013, Aeroflex Gaisler
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; if not, write to the Free Software
--  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

--Adapted for the use on the LEON3 Usecase at the JRZ VECS - FH Technikum Wien (c) 2013/2014

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
library grlib, techmap;
use grlib.amba.all;
use grlib.devices.all;
use grlib.stdlib.all;
use techmap.gencomp.all;
library gaisler;
use gaisler.memctrl.all;
use gaisler.leon3.all;
use gaisler.uart.all;
use gaisler.misc.all;
use gaisler.i2c.all;
use gaisler.jtag.all;
-- pragma translate_off
use gaisler.sim.all;
-- pragma translate_on

library esa;
use esa.memoryctrl.all;
use work.config.all;
use work.mypackage.all; --contains type

use work.assertion_checkers_p.all; --contains assertion checkers

entity leon3mp is
  generic (
    fabtech   : integer := CFG_FABTECH;
    memtech   : integer := CFG_MEMTECH;
    padtech   : integer := CFG_PADTECH;
    clktech   : integer := CFG_CLKTECH;
    disas     : integer := CFG_DISAS;	-- Enable disassembly to console
    dbguart   : integer := CFG_DUART;	-- Print UART on console
    pclow     : integer := CFG_PCLOW;
    SIMULATION : integer := 0
  );
  port (
    resetn	: in  std_logic; --key[0]
    clock_50	: in  std_logic;
    err_o      : out std_logic_vector(3 downto 0);
    errorn	: out std_logic; --ledr[0], error from LEON3 DSU
    uart_txd  	: out std_logic;		-- DSU, uart tx data
    uart_rxd  	: in  std_logic;		-- DSU,uart rx data
    dsubre  	: in std_logic;  --key[1], used to put processor in debug mode.
    dsuact  	: out std_logic; --ledr[1]
    i2c_scl		: inout std_logic;
    i2c_sda		: inout std_logic;
    gpio_0_final        : out std_logic_vector(CFG_GRGPIO_WIDTH-1 downto 0); 	-- multiplexed I/O port 0
    gpio_1        : out std_logic_vector(CFG_GRGPIO2_WIDTH-1 downto 0); 	-- I/O port 1
    fault_injection_buttons_async : in std_logic_vector(1 downto 0);
    ledg :out std_logic_vector(3 downto 0);
    dipsw_n : in std_logic_vector(3 downto 0); -- inject faults
    sw      	: in std_logic_vector(0 to 2) := "000" -- switch for jumper configuratioin, connects debug uart when sw(0) = '0'
    );
end;

architecture rtl of leon3mp is

-- Begin TODO: Move to config package -------
constant CFG_FAULT_INJECTION : integer range 0 to 1 := 1;
constant CFG_ASSERTION_COLLECTOR :integer range 0 to 1 := 1;
-- End TODO ---------------------------------


signal vcc, gnd   : std_logic_vector(4 downto 0);


--AMBA bus standard interface signals--
signal apbi  : apb_slv_in_type;
signal apbo  : apb_slv_out_vector := (others => apb_none);
signal ahbsi : ahb_slv_in_type;
signal ahbso : ahb_slv_out_vector := (others => ahbs_none);
signal ahbmi : ahb_mst_in_type;
signal ahbmo : ahb_mst_out_vector := (others => ahbm_none);


signal clkm, rstn, rstraw, pciclk, sdclkl, lclk, rst : std_logic;
--for  reset after fatal error
signal ext_rst_n_synced :std_logic;
signal self_rst :std_logic;
signal self_rst_r, self_rst_2r :std_logic;
signal self_reset :std_logic := '1';
signal self_rst_trig_start, self_rst_trig_end :std_logic;
-- end for  reset after fatal error

signal cgi : clkgen_in_type;
signal cgo : clkgen_out_type;

signal u1i, dui : uart_in_type;
signal u1o, duo : uart_out_type := (
  rtsn   	=> '0',
  txd   	   => '0',
  scaler	   => (others => '0'),
  txen     	=> '0',
  flow   	=> '0',
  rxen     	=> '0');


signal irqi : irq_in_vector(0 to CFG_NCPU-1);
signal irqo : irq_out_vector(0 to CFG_NCPU-1);

signal dbgi : l3_debug_in_vector(0 to CFG_NCPU-1) := (others => (
	
	dsuen   => '0',
	denable => '0',
	dbreak  => '0',  -- debug break-in
   step    => '0',  -- single step    
   halt    => '0',  -- halt processor
   reset   => '0',  -- reset processor
   dwrite  => '0',  -- read/write
   daddr   => (others => '0'), -- diagnostic address
   ddata   => (others => '0'), -- diagnostic data
   btrapa  => '0',	   -- break on IU trap
   btrape  => '0',	-- break on IU trap
   berror  => '0',	-- break on IU error mode
   bwatch  => '0',	-- break on IU watchpoint
   bsoft   => '0',	-- break on software breakpoint (TA 1)
   tenable => '0',
   timer   => (others => '0')));
			
signal dbgo : l3_debug_out_vector(0 to CFG_NCPU-1);

signal dsui : dsu_in_type;
signal dsuo : dsu_out_type; 

signal stati : ahbstat_in_type;

signal gpti : gptimer_in_type;

signal gpioi_0, gpioi_1 : gpio_in_type;
signal gpioo_0, gpioo_1 : gpio_out_type;

signal gpio_0      :std_logic_vector(CFG_GRGPIO_WIDTH-1 downto 0);

signal dsubren : std_logic;

signal tck, tms, tdi, tdo : std_logic;

signal fpi : grfpu_in_vector_type;
signal fpo : grfpu_out_vector_type;



constant BOARD_FREQ : integer := 50000;	-- Board frequency in KHz, used in clkgen
constant CPU_FREQ : integer := BOARD_FREQ * CFG_CLKMUL / CFG_CLKDIV;  -- cpu frequency in KHz (current 50Mhz)
constant IOAEN : integer := 1;
constant BUS_FREQ_KHZ :integer := BOARD_FREQ;

signal i2ci : i2c_in_type;
signal i2co : i2c_out_type;


attribute syn_keep : boolean;
attribute syn_preserve : boolean;
attribute keep : boolean;

signal aresetp : std_ulogic;

signal hready_b : std_logic_vector(15 downto 0);
signal err : std_logic_vector(3 downto 0);

--begin fault injection
signal s_fault_injection_button_sync : std_logic_vector(1 downto 0);
signal trigger_fault :std_logic_vector(3 downto 0);
constant FI_SWITCH_ON : std_logic := work.fault_injection_p.C_BUTTON_PRESSED;

constant TRIGGER_SUCCESSIVE_INTERRUPT :integer := 0;
constant TRIGGER_CTSN :integer := 1;
constant TRIGGER_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW : integer := 2;
constant TRIGGER_RULE5_ADDR_STABLE_DURING_TRANSITION : integer := 3;



	--uart ctsn condition check
	signal u1o_txd_r :std_logic_vector(1 downto 0);


	--fault containing signals
	signal my_ahbmo : ahb_mst_out_vector := (others => ahbm_none);
	signal my_apbi : apb_slv_in_type;
	signal fi_u1o_rtsn :std_logic;
	signal fi_address :std_logic_vector((2**4)-1 downto 0);
	
--end fault injection

--begin assertion checkers
signal ass_pending : std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
signal ass_done    : std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
signal ass_failed  : std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
signal fatal_ass_fail :std_logic_vector(ASS_STAT_WIDTH-1 downto 0);
signal assertions_status : assertion_status_t;
--end assertion checkers

begin



ass_pending(ASS_STAT_WIDTH-1 downto 8) <= (others =>'0');
ass_done(ASS_STAT_WIDTH-1 downto 8) <= (others =>'0');
ass_failed(ASS_STAT_WIDTH-1 downto 8) <= (others =>'0');

----------------------------------------------------------------------
---  Reset and Clock generation  -------------------------------------
----------------------------------------------------------------------

  vcc <= (others => '1'); gnd <= (others => '0');
  cgi.pllctrl <= "00"; cgi.pllrst <= rstraw;
  aresetp <= not(cgi.pllrst);

  clk_pad : clkpad generic map (tech => padtech) port map (clock_50, lclk); 

  clkgen0 : entity work.clkgen_de2
	generic map (clk_mul => CFG_CLKMUL, clk_div => CFG_CLKDIV, 
		clk_freq => BOARD_FREQ, sdramen => 0)
	port map (inclk0 => lclk, c0 => clkm, c0_2x => open, e0 => open,
		  locked => cgo.clklock, areset => aresetp);

	--simulation only
	--self_reset <= '0' after 1 ms, '1' after 1000020 ns ;--20 ns later

	
  resetn_pad : inpad generic map (tech => padtech) port map (resetn, rst); 
  rst0 : rstgen			-- reset generator (reset is active LOW)
    port map (rst, clkm, cgo.clklock, ext_rst_n_synced, rstraw);
	rstn <= ext_rst_n_synced and self_rst;
    
-----------------------------------------------------------------------
---  Fault injection: button sync/debounce ----------------------------
-----------------------------------------------------------------------

-- no use right now
err_o(3 downto 0) <= "1111";-- & s_fault_injection_button_sync;

fi_button_sync: if CFG_FAULT_INJECTION = 1 generate

	fpga :if SIMULATION = 0 generate
    fibs_i : work.fault_injection_p.fault_injection_button_sync
        generic map(
            G_DEBOUNCE_COUNTER_WIDTH => work.fault_injection_p.numbits(BOARD_FREQ),
                                            -- desired frequency: 1kHz
                                            -- 50MHz/65536 = 762.94Hz
            G_DEBOUNCE_SIGNAL_WIDTH  => 2,  -- 2 FPGA user buttons
            G_SINGLE_SHOT_MASK       => 1   -- All inputs non-singleshot
        )
        port map(
            clk_i => clkm,
            reset_n_i => rstn,
            async_i => fault_injection_buttons_async,
            sync_o  => s_fault_injection_button_sync
        );
     end generate;
     
    sim: if SIMULATION = 1 generate
			s_fault_injection_button_sync <= fault_injection_buttons_async;
    end generate;
    
end generate;

nofi_button_sync: if CFG_FAULT_INJECTION = 0 generate
    s_fault_injection_button_sync <= (others => not(work.fault_injection_p.C_BUTTON_PRESSED));
end generate;


-----------------------------------------------------------------------
---  Fault injection: choose fault to inject ----------------------------
-----------------------------------------------------------------------

fi_decoder: if CFG_FAULT_INJECTION = 1 generate
		fi_decoder0: work.fault_injection_p.fault_injection_decoder
    generic map(
        G_ADDRESS_WIDTH => 4,
        G_INVERT_BCD_ADDRESS => true,
        G_INVERT_ONEHOT0_ADDRESS => false
    )
    port map(
        clk_i     => clkm,
        reset_n_i => rstn,
        next_address_i => s_fault_injection_button_sync(0), -- trigger address increment
				computed_address_o => ledg,
				computed_address_onehot0_o => fi_address
    );
end generate;

-----------------------------------------------------------------------
---  Fault injection: inject faults on ahb(0|1) ----------------------------
-----------------------------------------------------------------------


fi_button_sync1: if CFG_FAULT_INJECTION = 1 generate

	----
	--case timer interrupt successive cpu1
	----
	trigger_fault(TRIGGER_SUCCESSIVE_INTERRUPT) <= FI_SWITCH_ON when (
			my_ahbmo(1).haddr = X"80000318"
			and my_ahbmo(1).hwdata(4) = '1' and fi_address(TRIGGER_SUCCESSIVE_INTERRUPT) = '1'
			and s_fault_injection_button_sync(1) = work.fault_injection_p.C_BUTTON_PRESSED
			)
		else
			not (FI_SWITCH_ON);

  fi_timer_interrupt_successive_cpu1 : work.fault_injection_p.fault_injection_unit 
    generic map(
        G_ASYNC_TRIGGER           => false,
	G_DEBOUNCE_COUNTER_WIDTH  => 1,
        G_OPERATION               => work.fault_injection_p.OP_OVW,
        G_INJECT_SINGLE_SHOT      => true,
        G_RANDOMIZE               => false,
        G_DATA_WIDTH              => 32,
        G_FI_DATA                 => 16#4000031C#,
        G_LFSR_POLY               => 16#00000071#, 
	G_LFSR_WIDTH      => 1
    )
    port map(
        clk_i                => clkm,
        reset_n_i            => rstn,
        trigger_i            => trigger_fault(TRIGGER_SUCCESSIVE_INTERRUPT),
    
        data_i              => my_ahbmo(1).haddr,
        data_o              => ahbmo(1).haddr
    );

		ctsn_fi_cond_check: process (clkm,rstn)
		begin
			if(rstn = '0') then
				u1o_txd_r <= (others => '1');
			elsif rising_edge(clkm) then
				u1o_txd_r(0) <= u1o.txd;
				u1o_txd_r(1) <= u1o_txd_r(0);
			end if;
		end process;

		--trigger upon falling edge of txd
		trigger_fault(TRIGGER_CTSN) <= FI_SWITCH_ON when (
			u1o_txd_r(1) = '1' and  u1o_txd_r(0) = '0'
			and fi_address(TRIGGER_CTSN) = '1'
			and s_fault_injection_button_sync(1) = work.fault_injection_p.C_BUTTON_PRESSED
		)
		else
			not (FI_SWITCH_ON);

    
    fi_uart_obeys_ctsn : work.fault_injection_p.fault_injection_unit
    generic map(
        G_ASYNC_TRIGGER           => false,
				G_DEBOUNCE_COUNTER_WIDTH  => 1,
        G_OPERATION               => work.fault_injection_p.OP_OVW,
        G_INJECT_SINGLE_SHOT      => false, --has no meaning if ASYNC_TRIGGER is false
        G_RANDOMIZE               => false,
        G_DATA_WIDTH              => 1,
        G_FI_DATA                 => 16#0001#,
        G_LFSR_POLY               => 16#5555#,
				G_LFSR_WIDTH      => 1
    )
    port map(
        clk_i                => clkm,
        reset_n_i            => rstn,
        trigger_i            => trigger_fault(TRIGGER_CTSN),

        data_i(0)              => u1o.rtsn,
        data_o(0)              => fi_u1o_rtsn
    );

			
		trigger_fault(TRIGGER_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW) <= FI_SWITCH_ON when (
			ahbmi.hready = '1' and
			my_ahbmo(1).htrans = "00" and --idle
			fi_address(TRIGGER_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW) = '1' and
			s_fault_injection_button_sync(1) = work.fault_injection_p.C_BUTTON_PRESSED
			)
		else
			not (FI_SWITCH_ON);
    
     fi_rule22 : work.fault_injection_p.fault_injection_unit
     generic map(
         G_ASYNC_TRIGGER           => false,
 				G_DEBOUNCE_COUNTER_WIDTH  => 1,
         G_OPERATION               => work.fault_injection_p.OP_OVW,
         G_INJECT_SINGLE_SHOT      => false, --has no meaning if ASYNC_TRIGGER is false
         G_RANDOMIZE               => false,
         G_DATA_WIDTH              => 2,
         G_FI_DATA                 => 16#3#,-- HTRANS_SEQ
         G_LFSR_POLY               => 16#0#,
 				G_LFSR_WIDTH      => 1
     )
     port map(
         clk_i                => clkm,
         reset_n_i            => rstn,
         trigger_i            => trigger_fault(TRIGGER_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW),
 
         data_i              => my_ahbmo(1).htrans,
         data_o              => ahbmo(1).htrans
     );
		--fault injection up uart access
		trigger_fault(TRIGGER_RULE5_ADDR_STABLE_DURING_TRANSITION) <= FI_SWITCH_ON when (
			my_apbi.paddr /= X"00000000" and
			my_apbi.psel(1) = '1' and
			my_apbi.penable = '1' and
			fi_address(TRIGGER_RULE5_ADDR_STABLE_DURING_TRANSITION) = '1' and
			s_fault_injection_button_sync(1) = work.fault_injection_p.C_BUTTON_PRESSED
			)
		else
			not (FI_SWITCH_ON);

     fi_rule5_apb : work.fault_injection_p.fault_injection_unit
     generic map(
         G_ASYNC_TRIGGER           => false,
 				G_DEBOUNCE_COUNTER_WIDTH  => 1,
         G_OPERATION               => work.fault_injection_p.OP_OR,
         G_INJECT_SINGLE_SHOT      => false, --has no meaning if ASYNC_TRIGGER is false
         G_RANDOMIZE               => false,
         G_DATA_WIDTH              => 32,
         G_FI_DATA                 => 16#55555555#,
         G_LFSR_POLY               => 16#0#,
 				G_LFSR_WIDTH      => 1
     )
     port map(
         clk_i                => clkm,
         reset_n_i            => rstn,
         trigger_i            => trigger_fault(TRIGGER_RULE5_ADDR_STABLE_DURING_TRANSITION),

         data_i              => my_apbi.paddr,
         data_o              => apbi.paddr
     );

    
end generate;
	--amba ahb masters
  ahbmo(0) <= my_ahbmo(0);

  ahbmo(1).hbusreq <= my_ahbmo(1).hbusreq;	
  ahbmo(1).hlock <= my_ahbmo(1).hlock;	
  --ahbmo(1).htrans <= my_ahbmo(1).htrans;	
  ahbmo(1).hwrite <= my_ahbmo(1).hwrite;	
  ahbmo(1).hsize <= my_ahbmo(1).hsize;	
  ahbmo(1).hburst <= my_ahbmo(1).hburst;	
  ahbmo(1).hprot <= my_ahbmo(1).hprot;	
  ahbmo(1).hwdata <= my_ahbmo(1).hwdata;
  ahbmo(1).hirq   <= my_ahbmo(1).hirq; 	
  ahbmo(1).hconfig  <= my_ahbmo(1).hconfig;
  ahbmo(1).hindex <= my_ahbmo(1).hindex;

  more_than_2_cpus: if CFG_NCPU > 2 generate 
    ahbm_interconnect: for i in 2 to CFG_NCPU-1 generate
      ahbmo(i) <= my_ahbmo(i);
    end generate;
  end generate;

  --amba apb ctrl
	apbi.psel <= my_apbi.psel;
	--apbi.paddr <= my_apbi.paddr;
  apbi.penable <= my_apbi.penable;
  apbi.pwrite <= my_apbi.pwrite;
  apbi.pwdata <= my_apbi.pwdata;
  apbi.pirq <= my_apbi.pirq;
  apbi.testen <= my_apbi.testen;
  apbi.testrst <= my_apbi.testrst;
	apbi.scanen <= my_apbi.scanen;
	apbi.testoen <= my_apbi.testoen;
	apbi.testin <= my_apbi.testin;


nofi_button_sync1: if CFG_FAULT_INJECTION = 0 generate
    s_fault_injection_button_sync <= (others => not(work.fault_injection_p.C_BUTTON_PRESSED));
    ahbmo(0).haddr <= my_ahbmo(1).haddr;
    apbi.paddr <= my_apbi.paddr;
    trigger_fault <= (others => not FI_SWITCH_ON);
end generate;


----------------------------------------------------------------------
---  AHB CONTROLLER --------------------------------------------------
----------------------------------------------------------------------

  ahb0 : ahbctrl 		-- AHB arbiter/multiplexer
  generic map (defmast => CFG_DEFMST, split => CFG_SPLIT, 
	rrobin => CFG_RROBIN, ioaddr => CFG_AHBIO, ioen => IOAEN,
	nahbm => CFG_NCPU+CFG_AHB_UART+CFG_AHB_JTAG,
	devid => ALTERA_DE2, nahbs => 8) 

  port map (rstn, clkm, ahbmi, ahbmo, ahbsi, ahbso);

  
  
  --------------------------------------------------------------------
  -- BEGIN CHECK
  
--    err_o <= err;
--  
--  p_check : process(clkm,rstn)
--  begin
--    if rstn = '0' then
--        hready_b <= (others => '1');
--        err <= "1111";
--    elsif rising_edge(clkm) then
--        if(apbi.penable = '1') then
--            err <= not(apbi.psel(12)&apbi.psel(9)&apbi.psel(3)&apbi.psel(2));
--        end if;
--    end if;
--  end process p_check;
  -- END CHECK
  --------------------------------------------------------------------
  
  
  ---------------------------------------------------------------------
  -- R6 Check
  
--  r6_i : r6_addr_incr_check_1reg
--    port map(
--        clk => clkm,
--        res_n => rstn,
--        htrans_i =>ahbmo(0).htrans,
--        hburst_i =>ahbmo(0).hburst,
--        
--        haddr_i  =>ahbmo(0).haddr,
--        hsize_i  =>ahbmo(0).hsize,
--        
--        res_o    => err_o(0)
--    );
  ---------------------------------------------------------------------
  
  
  
----------------------------------------------------------------------i
-----  LEON3 processor and DSU -----------------------------------------
----------------------------------------------------------------------




  cpu : for i in 0 to CFG_NCPU-1 generate
    nosh : if CFG_GRFPUSH = 0 generate    
      u0 : leon3s 		-- LEON3 processor      
      generic map (i, fabtech, memtech, CFG_NWIN, CFG_DSU, CFG_FPU*(1-CFG_GRFPUSH), CFG_V8, 
	0, CFG_MAC, pclow, CFG_NOTAG, CFG_NWP, CFG_ICEN, CFG_IREPL, CFG_ISETS, CFG_ILINE, 
	CFG_ISETSZ, CFG_ILOCK, CFG_DCEN, CFG_DREPL, CFG_DSETS, CFG_DLINE, CFG_DSETSZ,
	CFG_DLOCK, CFG_DSNOOP, CFG_ILRAMEN, CFG_ILRAMSZ, CFG_ILRAMADDR, CFG_DLRAMEN,
        CFG_DLRAMSZ, CFG_DLRAMADDR, CFG_MMUEN, CFG_ITLBNUM, CFG_DTLBNUM, CFG_TLB_TYPE, CFG_TLB_REP, 
        CFG_LDDEL, disas, CFG_ITBSZ, CFG_PWD, CFG_SVT, CFG_RSTADDR, CFG_NCPU-1,
	0, 0, CFG_MMU_PAGE, CFG_BP)
      port map (clkm, rstn, ahbmi, my_ahbmo(i), ahbsi, ahbso, 
    		irqi(i), irqo(i), dbgi(i), dbgo(i));
    end generate;
  end generate;

  ---ahb master (here cpu only) assertion checker(s)
  ahb_master_checker : for i in 0 to CFG_NCPU-1 generate
  ahbmchk: rule22_htrans_change_when_hready_low
	port map (
		clk_i  => clkm,
		reset_i   => rstn,
		-- assertion inputs --
		hready =>	ahbsi.hready,
		htrans => ahbmo(i).htrans,
		
		-- assertion status --
		pending_o => ass_pending(5+i),
		done_o => ass_done(5+i),
		failed_o => ass_failed(5+i)

	);
  end generate;
  
  
  sh : if CFG_GRFPUSH = 1 generate
    cpu : for i in 0 to CFG_NCPU-1 generate
      u0 : leon3sh 		-- LEON3 processor      
      generic map (i, fabtech, memtech, CFG_NWIN, CFG_DSU, CFG_FPU*(1-CFG_GRFPUSH), CFG_V8, 
	0, CFG_MAC, pclow, CFG_NOTAG, CFG_NWP, CFG_ICEN, CFG_IREPL, CFG_ISETS, CFG_ILINE, 
	CFG_ISETSZ, CFG_ILOCK, CFG_DCEN, CFG_DREPL, CFG_DSETS, CFG_DLINE, CFG_DSETSZ,
	CFG_DLOCK, CFG_DSNOOP, CFG_ILRAMEN, CFG_ILRAMSZ, CFG_ILRAMADDR, CFG_DLRAMEN,
        CFG_DLRAMSZ, CFG_DLRAMADDR, CFG_MMUEN, CFG_ITLBNUM, CFG_DTLBNUM, CFG_TLB_TYPE, CFG_TLB_REP, 
        CFG_LDDEL, disas, CFG_ITBSZ, CFG_PWD, CFG_SVT, CFG_RSTADDR, CFG_NCPU-1,
	0, 0, CFG_MMU_PAGE, CFG_BP)
      port map (clkm, rstn, ahbmi, my_ahbmo(i), ahbsi, ahbso, 
    		irqi(i), irqo(i), dbgi(i), dbgo(i), fpi(i), fpo(i));
    end generate;

    grfpush0 : grfpushwx generic map ((CFG_FPU-1), CFG_NCPU, fabtech)
      port map (clkm, rstn, fpi, fpo);
    
  end generate;
  --ledr[0] lit when leon 3 debugvector signals error
  errorn_pad : odpad generic map (tech => padtech) port map (errorn, dbgo(0).error);
  
  dsugen : if CFG_DSU = 1 generate
    dsu0 : dsu3			-- LEON3 Debug Support Unit (slave)
    generic map (hindex => 2, haddr => 16#900#, hmask => 16#F00#, 
       ncpu => CFG_NCPU, tbits => 30, tech => memtech, irq => 0, kbytes => CFG_ATBSZ)
    port map (rstn, clkm, ahbmi, ahbsi, ahbso(2), dbgo, dbgi, dsui, dsuo);
    dsui.enable <= '1';
    dsubre_pad : inpad generic map (tech => padtech) port map (dsubre, dsubren);
    dsuact_pad : outpad generic map (tech => padtech) port map (dsuact, dsuo.active); --ledr[1] is lit in debug mode.
    dsui.break <= not dsubren;
  end generate; 
  nodsu : if CFG_DSU = 0 generate 
    ahbso(2) <= ahbs_none; dsuo.tstop <= '0'; dsuo.active <= '0'; --no timer freeze, no light.
  end generate;

  dcomgen : if CFG_AHB_UART = 1 generate
    dcom0: ahbuart		-- Debug UART
    generic map (hindex => CFG_NCPU, pindex => 7, paddr => 7)
    port map (rstn, clkm, dui, duo, apbi, apbo(7), ahbmi, ahbmo(CFG_NCPU));  
      dui.rxd <= uart_rxd; -- when sw(0) = '0' else '1';
  end generate;
  
  ahbjtaggen0 :if CFG_AHB_JTAG = 1 generate
    ahbjtag0 : ahbjtag generic map(tech => fabtech, hindex => CFG_NCPU+CFG_AHB_UART)
      port map(rstn, clkm, tck, tms, tdi, tdo, ahbmi, ahbmo(CFG_NCPU+CFG_AHB_UART),
               open, open, open, open, open, open, open, gnd(0));
  end generate;
  
----------------------------------------------------------------------
--- On Chip Memories ----------------------------------------------
----------------------------------------------------------------------

--CFG_AHBRPIPE
  ahbramgen :if  CFG_AHBRAMEN = 1 generate
    ahbram0 : ahbram generic map (hindex => 3, haddr => CFG_AHBRADDR, 
	tech => CFG_MEMTECH, kbytes => CFG_AHBRSZ, pipe => CFG_AHBRPIPE)
      port map ( rstn, clkm, ahbsi, ahbso(3));
  end generate;
  noahbramgen : if CFG_AHBRAMEN = 0 generate
     ahbso(3) <= ahbs_none;
  end generate;

  ahbpromgen : if CFG_AHBROMEN /= 0 generate
    ahbprom0 : entity work.ahbrom
      generic map (hindex => 4, haddr => CFG_AHBRODDR, pipe => CFG_AHBROPIP)
      port map ( rstn, clkm, ahbsi, ahbso(4));
  end generate;
  noahbpromgen : if CFG_AHBROMEN = 0 generate
     ahbso(4) <= ahbs_none;
  end generate;


----------------------------------------------------------------------
---  APB Bridge and various periherals -------------------------------
----------------------------------------------------------------------

  apb0 : apbctrl				-- AHB/APB bridge
  generic map (hindex => 1, haddr => CFG_APBADDR)
  port map (rstn, clkm, ahbsi, ahbso(1), my_apbi, apbo );
  
	--assertion checker
	--
	i_rule5_addr_stable_during_transition_safety: timer_interrupt_successive_cpu1
	port map (
		clk_i     => clkm,
		reset_i   => rstn,
		haddr_0	=> ahbmo(0).haddr,
		hwrite_0 => ahbmo(0).hwrite,
		hwdata_0 => ahbmo(0).hwdata(7),
		pirq => apbo(11).pirq(11),
		pending_o => ass_pending(0),
		done_o => ass_done(0),
		failed_o => ass_failed(0)
	);

	
	i_rule5_addr_stable_during_transition:  rule5_addr_stable_during_transition
	port map(
		clk_i  => clkm,
		reset_ni =>  rstn,
		-- assertion inputs --
		psel_i => apbi.psel,
		penable_i => apbi.penable,
		paddr_i => apbi.paddr,
		pwrite_i => apbi.pwrite,
		pwdata_i => apbi.pwdata,

		-- assertion status --
		pending_o => ass_pending(7),
		done_o => ass_done(7),
		failed_o => ass_failed(7)
	);

	
	--end checker
  
----------------------------------------------------------------------------------------

 ass_coll : if CFG_ASSERTION_COLLECTOR /= 0 generate
 
      process(ass_done,ass_pending,ass_failed)
      begin

      --assertions_status(i).pending <= ass_pending(i);
      --assertions_status(i).done <= ass_done(i);
				for i in 0 to ASS_STAT_WIDTH-1 loop
					assertions_status(i).failed <= ass_failed(i);
				end loop;
      end process;
 
    collector:  assertion_collector
	generic map(
        pindex   => 11,
        paddr    => 11,
        pirq     => (11)
        )
	port map(
        rst    => rstn,
        clk    =>  clkm,
        apbi   => apbi,
        apbo   => apbo(11),
        self_reset_o => self_reset,
        fatal_fail_o => fatal_ass_fail,
        refeed_self_reset(0) => self_rst_trig_start,
        refeed_self_reset(1) => self_rst_trig_end,
        ass_stat => assertions_status,
        resval => (others => '0')
        );

fpga_self_reset_sim :if SIMULATION = 1 generate
	i_self_reset_generator : wait_counter
	generic map(
	  G_LOWER_BOUND  =>10,
	  G_UPPER_BOUND => 50000)--about 1ms at 50MHz
	port map(
		clk_i  =>  clkm,
		reset_i  =>  self_reset, --unless triggered(1) this signal is zero ->
														-- this starts the reset process
		enable_i => '1',
		lower_thr_o => self_rst_trig_start,
		upper_thr_o => self_rst_trig_end
	);
end generate;

fpga_self_reset :if SIMULATION = 0 generate
	i_self_reset_generator : wait_counter
		generic map(
			G_LOWER_BOUND  =>10,
			G_UPPER_BOUND => 500000000)--about 10s at 50MHz
		port map(
			clk_i  =>  clkm,
			reset_i  =>  self_reset, --unless triggered(1) this signal is zero ->
														-- this starts the reset process
			enable_i => '1',
			lower_thr_o => self_rst_trig_start,
			upper_thr_o => self_rst_trig_end
	);
end generate;

  --also fatal assertions must light up the leds
  process(fatal_ass_fail,gpio_0)
  begin
		for i in 0 to ASS_STAT_WIDTH-1 loop
			if fatal_ass_fail(i) = '1' then
				gpio_0_final(i) <= '1';
			else
				gpio_0_final(i) <= gpio_0(i);
			end if;
		end loop;
  end process;
	
	self_rst <= '0' when self_rst_trig_start = '1' and self_rst_trig_end = '0' else '1';

 
  end generate;
  noass_coll : if CFG_ASSERTION_COLLECTOR = 0 generate 
    apbo(11) <= apb_none;
    self_reset <= '1';
  end generate;
  
  ua1 : if CFG_UART1_ENABLE /= 0 generate
    uart1 : apbuart			-- UART 1
    generic map (pindex => 1, paddr => 1,  pirq => 2, console => dbguart, flow => 1,
	fifosize => CFG_UART1_FIFO)
    port map (rstn, clkm, apbi, apbo(1), u1i, u1o);
    u1i.extclk <= '0';
    u1i.rxd <= u1o.txd;
    u1i.ctsn <= fi_u1o_rtsn;
    --u1i.rxd <= '1' when sw(0) = '0' else uart_rxd;

    --begin asserion checker
    uart_obeys_ctsn_0: uart_obeys_ctsn
	port map(
		clk_i     => clkm,
		reset_i =>  rstn,
		-- assertion inputs --
		txd => u1o.txd,
		ctsn =>  fi_u1o_rtsn,
		-- assertion status --
		pending_o => ass_pending(4),
		done_o => ass_done(4),
		failed_o => ass_failed(4)
	);

    --end assertion checker
    
  end generate;
  uart_txd <= duo.txd; -- when sw(0) = '0' else u1o.txd;
  noua0 : if CFG_UART1_ENABLE = 0 generate apbo(1) <= apb_none; end generate;

  irqctrl : if CFG_IRQ3_ENABLE /= 0 generate
    irqctrl0 : irqmp			-- interrupt controller
    generic map (pindex => 2, paddr => 2, ncpu => CFG_NCPU)
    port map (rstn, clkm, apbi, apbo(2), irqo, irqi);
  end generate;
  irq3 : if CFG_IRQ3_ENABLE = 0 generate
    x : for i in 0 to CFG_NCPU-1 generate irqi(i).irl <= "0000"; end generate;
    apbo(2) <= apb_none;
  end generate;
 
  --Timer unit, generates interrupts when a timer underflow.
  gpt : if CFG_GPT_ENABLE /= 0 generate
    timer0 : gptimer 			-- timer unit
    generic map (pindex => 3, paddr => 3, pirq => CFG_GPT_IRQ, 
	sepirq => CFG_GPT_SEPIRQ, sbits => CFG_GPT_SW, ntimers => CFG_GPT_NTIM, 
	nbits => CFG_GPT_TW)
    port map (rstn, clkm, apbi, apbo(3), gpti, open);
    gpti.dhalt <= dsuo.tstop; gpti.extclk <= '0';
    
    --begin assertion_checker
    --timer_interrupt_never_cpu0_0 : timer_interrupt_never_cpu0
    --  port map(
		--		clk_i => clkm,
		--		reset_i => rstn,
		--		haddr => ahbmo(0).haddr,
		--		hwrite => ahbmo(0).hwrite,
		--		hwdata => ahbmo(0).hwdata(8),
		--		pirq => apbo(3).pirq(8),
		--		pending_o => ass_pending(0),
		--		done_o => ass_done(0),
		--		failed_o => ass_failed(0)
    -- );
      
      timer_interrupt_cpu1_0 : timer_interrupt_cpu1
      port map(
				clk_i => clkm,
				reset_i => rstn,
				haddr_0 => ahbmo(0).haddr,
				hwrite_0 => ahbmo(0).hwrite,
				hwdata_0 => ahbmo(0).hwdata(8),
				haddr_1 => ahbmo(1).haddr,
				hwrite_1 => ahbmo(1).hwrite,
				hwdata_1 => ahbmo(1).hwdata(4),
				pirq => apbo(3).pirq(8),
				pending_o => ass_pending(1),
				done_o => ass_done(1),
				failed_o => ass_failed(1)
      );
      
    timer_interrupt_successive_cpu1_0 : timer_interrupt_successive_cpu1
	port map (
		clk_i     => clkm,
		reset_i   => rstn,
		haddr_0	=> ahbmo(1).haddr,
		hwrite_0 => ahbmo(1).hwrite,
		hwdata_0 => ahbmo(1).hwdata(4),
		pirq => apbo(3).pirq(8),	
		pending_o => ass_pending(2),
		done_o => ass_done(2),
		failed_o => ass_failed(2)
	);

      
      
    --end assertion_checker
    
  end generate;
  notim : if CFG_GPT_ENABLE = 0 generate apbo(3) <= apb_none; end generate;

  gpio0 : if CFG_GRGPIO_ENABLE /= 0 generate     -- GR GPIO0 unit
    grgpio0: grgpio
      generic map( pindex => 9, paddr => 9, imask => CFG_GRGPIO_IMASK, nbits => CFG_GRGPIO_WIDTH, bypass => 0)
      port map( rstn, clkm, apbi, apbo(9), gpioi_0, gpioo_0);
      gpioi_0.din <= (others => '0');
      gpio_0 <=  gpioo_0.dout;
      connect_0 : for i in 0 to CFG_GRGPIO_WIDTH-1 generate
	gpio_0(i) <=  gpioo_0.dout(i);
      end generate;
      --pio_pads : for i in 0 to CFG_GRGPIO_WIDTH-1 generate
      --  pio_pad : iopad generic map (tech => padtech)
      --      port map (gpio_0(i), gpioo_0.dout(i), gpioo_0.oen(i));--, gpioi_0.din(i));
      --end generate;
  end generate;
  nogpio0: if CFG_GRGPIO_ENABLE = 0 generate apbo(9) <= apb_none; end generate;

  gpio1 : if CFG_GRGPIO2_ENABLE /= 0 generate     -- GR GPIO1 unit
    grgpio1: grgpio
      generic map( pindex => 10, paddr => 10, imask => CFG_GRGPIO2_IMASK, nbits => CFG_GRGPIO2_WIDTH, bypass => 0)
      port map( rstn, clkm, apbi, apbo(10), gpioi_1, gpioo_1);
      gpioi_1.din <= (others => '0');
      connect : for i in 0 to CFG_GRGPIO_WIDTH-1 generate
	gpio_1(i) <=  gpioo_1.dout(i);
      end generate;
      
      --pio_pads : for i in 0 to CFG_GRGPIO2_WIDTH-1 generate
      --  pio_pad : iopad generic map (tech => padtech)
      --      port map (gpio_1(i), gpioo_1.dout(i), gpioo_1.oen(i));--, gpioi_1.din(i));
      --end generate;
  end generate;
  nogpio1: if CFG_GRGPIO2_ENABLE = 0 generate apbo(10) <= apb_none; end generate;
  
  
  i2cmst0: if CFG_I2C_ENABLE /= 0 generate --i2c master
    i2c0: i2cmst generic map (pindex => 12, paddr => 12, pmask => 16#FFF#, 
				pirq => 12, filter => (BUS_FREQ_KHZ*5+50000)/100000+1)
		  port map (rstn, clkm, apbi, apbo(12), i2ci, i2co);
    -- Using bi-directional pads:
    i2c_scl_pad : iopad generic map (tech => padtech)
			port map (i2c_scl, i2co.scl, i2co.scloen, i2ci.scl);
    i2c_sda_pad : iopad generic map (tech => padtech)
			port map (i2c_sda, i2co.sda, i2co.sdaoen, i2ci.sda);

  --begin assertion checker
    i2c_liveness_0 : i2c_liveness 
	  port map (
		clk_i     => clkm,
		reset_i  =>  rstn,
		-- assertion inputs --
		scl =>  i2ci.scl,
		sda => i2ci.sda,	
			
		-- assertion status --
		pending_o => ass_pending(3),
		done_o => ass_done(3),
		failed_o => ass_failed(3)
	
	);
  
  
  --end assertion checker
			
			
  end generate;
  noi2cmst0: if CFG_I2C_ENABLE = 0 generate apbo(12) <= apb_none; end generate;
--   ahbs : if CFG_AHBSTAT = 1 generate	-- AHB status register
--     stati.cerror(1 to NAHBSLV-1) <= (others => '0');
--     stati.cerror(0) <= memo.ce; --connect as many fault tolerans units as specified by nftslv generic.
--     ahbstat0 : ahbstat generic map (pindex => 15, paddr => 15, pirq => 1,
-- 	nftslv => CFG_AHBSTATN)
--       port map (rstn, clkm, ahbmi, ahbsi, stati, apbi, apbo(15));
--   end generate;
--   nop2 : if CFG_AHBSTAT = 0 generate apbo(15) <= apb_none; end generate;


-----------------------------------------------------------------------
---  Drive unused bus elements  ---------------------------------------
-----------------------------------------------------------------------

--  nam1 : for i in (CFG_NCPU+CFG_AHB_UART+log2x(CFG_PCI)+CFG_AHB_JTAG) to NAHBMST-1 generate
--    ahbmo(i) <= ahbm_none;
--  end generate;
--  nam2 : if CFG_PCI > 1 generate
--    ahbmo(CFG_NCPU+CFG_AHB_UART+CFG_AHB_JTAG+log2x(CFG_PCI)-1) <= ahbm_none;
--  end generate;
--  nap0 : for i in 11 to NAPBSLV-1 generate apbo(i) <= apb_none; end generate;
--  apbo(6) <= apb_none;

-----------------------------------------------------------------------
---  Test report module  ----------------------------------------------
-----------------------------------------------------------------------

-- pragma translate_off

  test0 : ahbrep generic map (hindex => 7, haddr => 16#200#)
   port map (rstn, clkm, ahbsi, ahbso(7));

-- pragma translate_on
-----------------------------------------------------------------------
---  Boot message  ----------------------------------------------------
-----------------------------------------------------------------------

-- pragma translate_off
  x : report_design
  generic map (
   msg1 => "LEON3 Altera Cyclone V Design",
   fabtech => tech_table(fabtech), memtech => tech_table(memtech),
   mdel => 1
  );
-- pragma translate_on
end;