/**
 * main.c
 *
 * Created on:           Oct. 24, 2013
 * Adapted for LEON3 on: Nov. 19, 2013
 *     Author: Christian, Matthias
 *
 * Demonstration Software for the LEON3 usecase
 *
 * Periodically reads the I2C RTC and updates the LEDs on the board
 * Assertion checkers generate an interrupt if an assertion failed the corresponding
 * green led is lit.
 * led id: assertion failed
 * 0: RULE5_SAFETY_ASSERTION
 * 1: TIMER_IRQ_ASSIGNED_CPU1
 * 2: TIMER_IRQ_SUCCESSIVE_CPU1
 * 3: I2C_LIVENESS
 * 4: UART_OBEYS_CTSN
 * 5: RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW_CPU_1
 * 6: RULE5_ADDR_STABLE_DURING_TRANSITION
 *
 * Moreover rule5 and rule22 are flagged fatal causing the whole system to reset itself upon failing
 * 
 * CPU0: initialize hardware and setup life beacon (toggles LED periodically)
 * CPU1: read rtc and display on LEDs
 */

#include "inc_usecase_rtc/includes.h"
#include <irq.h>
#include <uart.h>

#ifdef DEBUG_TRAP_HANDLER
volatile uint32_t panic_id;
#endif


uint32_t cpuid;

GPIO_t *gpio0_dev = (GPIO_t *)GPIO0_BASE_ADDR;
GPIO_t *gpio1_dev = (GPIO_t *)GPIO1_BASE_ADDR;
TIMER_t *timer0_dev = (TIMER_t *)TIMER0_BASE_ADDR;
I2C_SLV_t *i2c_rtc_slv = (I2C_SLV_t *) RTC_BASE_ADDR;
I2C_SLV_t *i2c_lcd_slv = (I2C_SLV_t *) LCD_BASE_ADDR;
UART_t *uart_dev = (UART_t *) UART_BASE_ADDR;

struct time my_time;

void init_hw(struct time * set_time) {

	#ifdef IN_FPGA
	/*configure the assertion capture peripheral and activate the given assertions*/
	enable_assertion_capture(
		(TIMER_IRQ_SUCCESSIVE_CPU1|
		UART_OBEYS_CTSN|
		RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW_CPU_1|
		RULE5_ADDR_STABLE_DURING_TRANSITION)
		,
		(FATAL_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW_CPU_1|
		FATAL_RULE5_ADDR_STABLE_DURING_TRANSITION)
	);
	#else
	enable_assertion_capture(
		(TIMER_IRQ_SUCCESSIVE_CPU1|
		UART_OBEYS_CTSN|
		RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW_CPU_1|
		RULE5_ADDR_STABLE_DURING_TRANSITION)
		,
		(FATAL_RULE22_HTRANS_CHANGE_WHEN_HREADY_LOW_CPU_1|
		FATAL_RULE5_ADDR_STABLE_DURING_TRANSITION)
	);
	#endif

	gpio0_dev->dir |= LED_MASK_10; /* Set 10 LSBs of GPIO to output (For HSMC LEDs) */
	gpio0_dev->output = 0;

	gpio1_dev->dir |= LED_MASK_10; /* Set 10 LSBs of GPIO to output (For HSMC LEDs) */
	gpio1_dev->output = 0;

	#ifdef IN_FPGA
	init_i2c(i2c_rtc_slv,I2C_SPEED_50kHz);
	
	init_display(i2c_lcd_slv);

	init_rtc(i2c_rtc_slv,set_time);
	#endif
	 
	uart_init(uart_dev);

}


/*toggle green led, id:9*/
void timer_isr(uint32_t id) {

	timer0_dev->tmr1ctrl |= (1<<4);
	gpio0_dev->output ^= (1<<9);
}

#ifdef DEBUG_TRAP_HANDLER
/*system panic handler see trap table*/
void _panic(uint32_t id){
	asm volatile ("set panic_id, %%l7\n" :/*regs out*/ : /*regs in*/:/*globber*/"%l7");
	asm volatile ("st %%i0,[%%l7]\n" : : :"%l7","%i0");

	gpio0_dev->output |= (1<<panic_id);
}
#endif

int main(void) {

	uint32_t cpuid;

	cpuid = get_cpuid();
	register_cpu(cpuid);

	if(cpuid == 0) {

		#ifdef IN_FPGA
		int i;
		#endif
		
		my_time.sec = DEC2BCD( 0);  /* Seconds      */
		my_time.min = DEC2BCD( 0);  /* Minutes      */
		my_time.hrs = DEC2BCD(12);  /* Hours        */
		my_time.dow =   6;            /* Day of Week  */
		my_time.dom = DEC2BCD(26);  /* Day in Month */
		my_time.month = DEC2BCD(10);  /* Month        */
		my_time.year = DEC2BCD(13); /* Year         */

		init_irqmp();
		init_hw(&my_time);

		#ifdef IN_FPGA
		display_puts(i2c_lcd_slv,"LEON/3 RESET    ",LINE_1);
		display_puts(i2c_lcd_slv,"***INITIALIZING ",LINE_2);
		#endif


		/*setup life beacon*/
		register_isr(TIMER_IRQ0,timer_isr);
		activate_interrupt(1, TIMER_IRQ0);/*on cpu 1*/

		/*setup timer 0, IN_FPGA blink frequency is  ~ 1Hz*/
		#ifdef IN_FPGA
		timer0_dev->presc_val = 1000;
		timer0_dev->presc_rel = 1000;

		timer0_dev->tmr1rel = 50000;
		#else
		timer0_dev->presc_val = 100;
		timer0_dev->presc_rel = 100;

		timer0_dev->tmr1rel = 500;
		#endif

		#ifdef IN_FPGA

		display_puts(i2c_lcd_slv,"LEON/3 RESET    ",LINE_1);
		display_puts(i2c_lcd_slv,"***INITIALIZING ",LINE_2);

		for(i = 0; i < 50; i++) delay_100us(200);

		display_puts(i2c_lcd_slv,"LEON/3 RUNNING  ",LINE_1);
		display_puts(i2c_lcd_slv,"***STANDARD MODE",LINE_2);

		#endif
		/*start 2nd cpu and wait until it's up*/
		start_cpu(1);
		/*start timer0  */
		timer0_dev->tmr1ctrl |= (1<<2);/*load reload value in timer counter register*/
		timer0_dev->tmr1ctrl = (1<<3) | (3<<0);/*set restart bit (automatic reload upon underflow) and start the timer*/

		for(;;) {
			#ifdef IN_FPGA
			delay_100us(1000);
			#else
			delay_100us(1);
			#endif
			/*send a meaningless bogus uart frame, to test ctsn assertion*/
			uart_send_byte(uart_dev,0x55);
			/*wait untill the byte is received and discard it*/
			(void) uart_recv_byte(uart_dev);
		}


	} else if(cpuid == 1) {

		for(;;) {
			#ifdef IN_FPGA
			delay_100us(1000);
			#else
			delay_100us(1);
			#endif
			#ifdef IN_FPGA
			/*display elapsed seconds modulo 10 on the red leds as onehot0 encoded*/
			read_rtc(i2c_rtc_slv,&my_time);
			display_seconds(my_time.sec);
			#endif
		}


	}

	/*should never be reached*/
	return 0;

}


